"""SQLAlchemy entity models"""

from datetime import datetime

from axion_lab_server.repos.models.base import Base
from sqlalchemy import (
    DateTime,
    Float,
    ForeignKey,
    Index,
    Integer,
    String,
    Text,
    UniqueConstraint,
)
from sqlalchemy.orm import Mapped, mapped_column, relationship


class Org(Base):
    """Organization entity"""

    __tablename__ = "orgs"

    org_id: Mapped[str] = mapped_column(String(26), primary_key=True)
    name: Mapped[str] = mapped_column(String(255), nullable=False)
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False
    )

    # Relationships
    projects: Mapped[list["Project"]] = relationship(
        "Project", back_populates="org", cascade="all, delete-orphan"
    )


class Project(Base):
    """Project entity"""

    __tablename__ = "projects"

    project_id: Mapped[str] = mapped_column(String(26), primary_key=True)
    org_id: Mapped[str] = mapped_column(
        String(26), ForeignKey("orgs.org_id"), nullable=False
    )
    name: Mapped[str] = mapped_column(String(255), nullable=False)
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False
    )

    # Relationships
    org: Mapped["Org"] = relationship("Org", back_populates="projects")
    batches: Mapped[list["Batch"]] = relationship(
        "Batch", back_populates="project", cascade="all, delete-orphan"
    )
    aggregations: Mapped[list["Aggregation"]] = relationship(
        "Aggregation", back_populates="project", cascade="all, delete-orphan"
    )

    __table_args__ = (Index("ix_projects_org_created", "org_id", "created_at"),)


class Batch(Base):
    """Batch entity (experiment condition grouping)"""

    __tablename__ = "batches"

    batch_id: Mapped[str] = mapped_column(String(26), primary_key=True)
    project_id: Mapped[str] = mapped_column(
        String(26), ForeignKey("projects.project_id"), nullable=False
    )
    name: Mapped[str] = mapped_column(String(255), nullable=False)
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False
    )

    # Relationships
    project: Mapped["Project"] = relationship("Project", back_populates="batches")
    runs: Mapped[list["Run"]] = relationship(
        "Run", back_populates="batch", cascade="all, delete-orphan"
    )
    dp_jobs: Mapped[list["DPJob"]] = relationship(
        "DPJob", back_populates="batch", cascade="all, delete-orphan"
    )

    __table_args__ = (Index("ix_batches_project_created", "project_id", "created_at"),)


class Run(Base):
    """Run entity (single experiment execution)"""

    __tablename__ = "runs"

    run_id: Mapped[str] = mapped_column(String(26), primary_key=True)
    batch_id: Mapped[str] = mapped_column(
        String(26), ForeignKey("batches.batch_id"), nullable=False
    )
    name: Mapped[str] = mapped_column(String(255), nullable=False)
    status: Mapped[str] = mapped_column(
        String(20), nullable=False, default="active"
    )  # active, garbage, archived
    tags_json: Mapped[str] = mapped_column(Text, nullable=False, default="[]")
    note: Mapped[str | None] = mapped_column(Text, nullable=True)
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False
    )
    updated_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False
    )

    # Relationships
    batch: Mapped["Batch"] = relationship("Batch", back_populates="runs")
    artifacts: Mapped[list["Artifact"]] = relationship(
        "Artifact", back_populates="run", cascade="all, delete-orphan"
    )
    quality_metrics: Mapped[list["QualityMetric"]] = relationship(
        "QualityMetric", back_populates="run", cascade="all, delete-orphan"
    )
    comparison_indicators: Mapped[list["ComparisonIndicator"]] = relationship(
        "ComparisonIndicator", back_populates="run", cascade="all, delete-orphan"
    )
    pins: Mapped[list["RunPin"]] = relationship(
        "RunPin", back_populates="run", cascade="all, delete-orphan"
    )
    aggregation_members: Mapped[list["AggregationMember"]] = relationship(
        "AggregationMember", back_populates="run", cascade="all, delete-orphan"
    )

    __table_args__ = (
        Index("ix_runs_batch_created", "batch_id", "created_at"),
        Index("ix_runs_batch_status_created", "batch_id", "status", "created_at"),
    )


class RunPin(Base):
    """Run pin entity (champion / user_selected)"""

    __tablename__ = "run_pins"

    pin_id: Mapped[str] = mapped_column(String(26), primary_key=True)
    run_id: Mapped[str] = mapped_column(
        String(26), ForeignKey("runs.run_id"), nullable=False
    )
    batch_id: Mapped[str] = mapped_column(
        String(26), ForeignKey("batches.batch_id"), nullable=False
    )
    pin_type: Mapped[str] = mapped_column(
        String(20), nullable=False
    )  # champion, user_selected
    pinned_by: Mapped[str | None] = mapped_column(String(255), nullable=True)
    pinned_at: Mapped[datetime] = mapped_column(DateTime(timezone=True), nullable=False)

    # Relationships
    run: Mapped["Run"] = relationship("Run", back_populates="pins")
    batch: Mapped["Batch"] = relationship("Batch")

    __table_args__ = (
        Index("ix_run_pins_batch_type", "batch_id", "pin_type", "pinned_at"),
        # Ensure only one champion per batch
        UniqueConstraint(
            "batch_id",
            "pin_type",
            name="uq_batch_champion",
            sqlite_on_conflict="REPLACE",
        ),
    )


class Artifact(Base):
    """Artifact entity (references or inline values)"""

    __tablename__ = "artifacts"

    artifact_id: Mapped[str] = mapped_column(String(26), primary_key=True)
    run_id: Mapped[str] = mapped_column(
        String(26), ForeignKey("runs.run_id"), nullable=False
    )
    kind: Mapped[str] = mapped_column(
        String(20), nullable=False
    )  # url, local, inline_text, inline_number, inline_json
    type: Mapped[str] = mapped_column(
        String(50), nullable=False
    )  # langfuse_trace, object_storage, note, evaluation, etc.
    label: Mapped[str] = mapped_column(String(255), nullable=False)
    payload_text: Mapped[str | None] = mapped_column(
        Text, nullable=True
    )  # for url, local, text, json
    payload_number: Mapped[float | None] = mapped_column(
        Float, nullable=True
    )  # for inline_number
    meta_json: Mapped[str] = mapped_column(Text, nullable=False, default="{}")
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False
    )

    # Relationships
    run: Mapped["Run"] = relationship("Run", back_populates="artifacts")

    __table_args__ = (
        Index("ix_artifacts_run_created", "run_id", "created_at"),
        Index("ix_artifacts_type_created", "type", "created_at"),
    )


class QualityMetric(Base):
    """Quality Metric entity (run quality indicators)"""

    __tablename__ = "quality_metrics"

    qm_id: Mapped[str] = mapped_column(String(26), primary_key=True)
    run_id: Mapped[str] = mapped_column(
        String(26), ForeignKey("runs.run_id"), nullable=False
    )
    key: Mapped[str] = mapped_column(String(100), nullable=False)
    value_json: Mapped[str] = mapped_column(Text, nullable=False)
    source: Mapped[str] = mapped_column(
        String(20), nullable=False
    )  # raw, derived, manual
    computed_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False
    )
    version: Mapped[int] = mapped_column(Integer, nullable=False, default=1)

    # Relationships
    run: Mapped["Run"] = relationship("Run", back_populates="quality_metrics")

    __table_args__ = (
        Index("ix_quality_metrics_run_computed", "run_id", "computed_at"),
        UniqueConstraint("run_id", "key", "version", name="uq_qm_run_key_version"),
    )


class ComparisonIndicator(Base):
    """Comparison Indicator entity (run comparison indicators)"""

    __tablename__ = "comparison_indicators"

    ci_id: Mapped[str] = mapped_column(String(26), primary_key=True)
    run_id: Mapped[str] = mapped_column(
        String(26), ForeignKey("runs.run_id"), nullable=False
    )
    key: Mapped[str] = mapped_column(String(100), nullable=False)
    value_json: Mapped[str] = mapped_column(Text, nullable=False)
    baseline_ref: Mapped[str | None] = mapped_column(String(26), nullable=True)
    computed_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False
    )
    version: Mapped[int] = mapped_column(Integer, nullable=False, default=1)

    # Relationships
    run: Mapped["Run"] = relationship("Run", back_populates="comparison_indicators")

    __table_args__ = (
        Index("ix_comparison_indicators_run_computed", "run_id", "computed_at"),
        UniqueConstraint("run_id", "key", "version", name="uq_ci_run_key_version"),
    )


class Aggregation(Base):
    """Aggregation entity (cross-batch metadata-based grouping).

    While a Batch groups Runs by experiment condition, an Aggregation groups
    Runs across batches (or within a batch) based on metadata keys.
    Typical use-cases:
    - ML epoch × method Loss/metric comparison
    - LLM model × agent implementation version comparison
    """

    __tablename__ = "aggregations"

    aggregation_id: Mapped[str] = mapped_column(String(26), primary_key=True)
    project_id: Mapped[str] = mapped_column(
        String(26), ForeignKey("projects.project_id"), nullable=False
    )
    name: Mapped[str] = mapped_column(String(255), nullable=False)
    description: Mapped[str | None] = mapped_column(Text, nullable=True)
    group_by_keys_json: Mapped[str] = mapped_column(
        Text, nullable=False, default="[]"
    )  # JSON list of metadata keys, e.g. ["epoch", "model_name"]
    filter_json: Mapped[str] = mapped_column(
        Text, nullable=False, default="{}"
    )  # JSON filter spec, e.g. {"model_name": ["gpt-4o", "claude-3"]}
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False
    )

    # Relationships
    project: Mapped["Project"] = relationship("Project", back_populates="aggregations")
    members: Mapped[list["AggregationMember"]] = relationship(
        "AggregationMember",
        back_populates="aggregation",
        cascade="all, delete-orphan",
    )

    __table_args__ = (
        Index("ix_aggregations_project_created", "project_id", "created_at"),
    )


class AggregationMember(Base):
    """Links a Run to an Aggregation with associated metadata values."""

    __tablename__ = "aggregation_members"

    member_id: Mapped[str] = mapped_column(String(26), primary_key=True)
    aggregation_id: Mapped[str] = mapped_column(
        String(26), ForeignKey("aggregations.aggregation_id"), nullable=False
    )
    run_id: Mapped[str] = mapped_column(
        String(26), ForeignKey("runs.run_id"), nullable=False
    )
    metadata_json: Mapped[str] = mapped_column(
        Text, nullable=False, default="{}"
    )  # e.g. {"epoch": 5, "model_name": "gpt-4o"}
    added_at: Mapped[datetime] = mapped_column(DateTime(timezone=True), nullable=False)

    # Relationships
    aggregation: Mapped["Aggregation"] = relationship(
        "Aggregation", back_populates="members"
    )
    run: Mapped["Run"] = relationship("Run", back_populates="aggregation_members")

    __table_args__ = (
        Index("ix_agg_members_agg_added", "aggregation_id", "added_at"),
        Index("ix_agg_members_run", "run_id"),
        UniqueConstraint("aggregation_id", "run_id", name="uq_aggregation_run"),
    )


class DPJob(Base):
    """DP Job entity (computation job tracking)"""

    __tablename__ = "dp_jobs"

    job_id: Mapped[str] = mapped_column(String(26), primary_key=True)
    batch_id: Mapped[str] = mapped_column(
        String(26), ForeignKey("batches.batch_id"), nullable=False
    )
    mode: Mapped[str] = mapped_column(
        String(20), nullable=False
    )  # active_only, include_garbage
    recompute: Mapped[int] = mapped_column(Integer, nullable=False, default=0)  # 0 or 1
    status: Mapped[str] = mapped_column(
        String(20), nullable=False, default="queued"
    )  # queued, running, succeeded, failed, canceled
    requested_by: Mapped[str | None] = mapped_column(String(255), nullable=True)
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False
    )
    started_at: Mapped[datetime | None] = mapped_column(
        DateTime(timezone=True), nullable=True
    )
    finished_at: Mapped[datetime | None] = mapped_column(
        DateTime(timezone=True), nullable=True
    )
    error_text: Mapped[str | None] = mapped_column(Text, nullable=True)

    # Relationships
    batch: Mapped["Batch"] = relationship("Batch", back_populates="dp_jobs")

    __table_args__ = (
        Index("ix_dp_jobs_batch_created", "batch_id", "created_at"),
        Index("ix_dp_jobs_status_created", "status", "created_at"),
    )



"""Pin API router"""

from typing import Annotated

from fastapi import APIRouter, HTTPException, Path, status

from axion_lab_server.apps.api.deps import BatchRepo, RunPath, RunPinRepo
from axion_lab_server.shared.domain import PinCreate, PinResponse, PinType

router = APIRouter(tags=["Pins"])


@router.post(
    "/runs/{run_id}/pins",
    response_model=PinResponse,
    status_code=status.HTTP_201_CREATED,
)
async def create_pin(
    data: PinCreate,
    run: RunPath,
    repo: RunPinRepo,
    batch_repo: BatchRepo,
) -> PinResponse:
    """Create a pin for a run (champion or user_selected)"""
    # Verify batch exists
    batch = await batch_repo.get_by_id(run.batch_id)
    if not batch:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Batch not found: {run.batch_id}",
        )

    # Check if pin already exists
    existing = await repo.get_by_run_and_type(run.run_id, data.pin_type)
    if existing:
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail=f"Pin already exists for run {run.run_id} with type {data.pin_type}",
        )

    pin = await repo.create(
        run_id=run.run_id,
        batch_id=run.batch_id,
        pin_type=data.pin_type,
    )
    return PinResponse.model_validate(pin)


PinTypeFromPath = Annotated[PinType, Path(...)]


@router.delete(
    "/runs/{run_id}/pins/{pin_type}",
    status_code=status.HTTP_204_NO_CONTENT,
)
async def delete_pin(
    run: RunPath,
    pin_type: PinTypeFromPath,
    repo: RunPinRepo,
) -> None:
    """Delete a pin from a run"""
    deleted = await repo.delete_by_run_and_type(run.run_id, pin_type)
    if not deleted:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Pin not found for run {run.run_id} with type {pin_type}",
        )



"""Batch service."""

from __future__ import annotations

from axion_lab.schemas import BatchCreate, BatchResponse, CursorPaginatedResponse
from axion_lab.services._http import HttpTransport


class BatchService:
    """CRUD operations for batches."""

    def __init__(self, http: HttpTransport) -> None:
        self._http = http

    def create(self, project_id: str, name: str) -> BatchResponse:
        """Create a new batch under a project."""
        data = self._http.post(
            f"/projects/{project_id}/batches",
            json=BatchCreate(name=name).model_dump(),
        )
        return BatchResponse.model_validate(data)

    def list(
        self,
        project_id: str,
        *,
        limit: int = 20,
        cursor: str | None = None,
    ) -> CursorPaginatedResponse[BatchResponse]:
        """List batches in a project."""
        params: dict[str, object] = {"limit": limit}
        if cursor:
            params["cursor"] = cursor
        data = self._http.get(f"/projects/{project_id}/batches", params=params)
        return CursorPaginatedResponse[BatchResponse].model_validate(data)

    def get(self, batch_id: str) -> BatchResponse:
        """Get a batch by ID."""
        data = self._http.get(f"/batches/{batch_id}")
        return BatchResponse.model_validate(data)



"""DP Job API router"""

from fastapi import APIRouter, BackgroundTasks, HTTPException, status

from axion_lab_server.apps.api.deps import BatchPath, DPJobRepo
from axion_lab_server.ops.dp import DPRunner
from axion_lab_server.shared.domain import DPJobCreate, DPJobResponse

router = APIRouter(tags=["DP Jobs"])


def _build_dp_job_response(job) -> DPJobResponse:
    """Build DP job response"""
    return DPJobResponse(
        jobId=job.job_id,
        batchId=job.batch_id,
        mode=job.mode,
        recompute=bool(job.recompute),
        status=job.status,
        requestedBy=job.requested_by,
        createdAt=job.created_at,
        startedAt=job.started_at,
        finishedAt=job.finished_at,
        errorText=job.error_text,
    )


@router.post(
    "/batches/{batch_id}/dp/compute",
    response_model=DPJobResponse,
    status_code=status.HTTP_201_CREATED,
)
async def create_dp_job(
    data: DPJobCreate,
    batch: BatchPath,
    repo: DPJobRepo,
    background_tasks: BackgroundTasks,
) -> DPJobResponse:
    """Create and trigger a DP computation job"""
    # Check if there's already a running job
    has_running = await repo.has_running_job(batch.batch_id)
    if has_running:
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail="A DP job is already running for this batch",
        )

    # Create job
    job = await repo.create(batch.batch_id, data)

    # Start DP runner in background
    background_tasks.add_task(
        DPRunner.run_job,
        job.job_id,
    )

    return _build_dp_job_response(job)


@router.get(
    "/dp/jobs/{job_id}",
    response_model=DPJobResponse,
)
async def get_dp_job(
    job_id: str,
    repo: DPJobRepo,
) -> DPJobResponse:
    """Get DP job status"""
    job = await repo.get_by_id(job_id)
    if not job:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"DP Job not found: {job_id}",
        )
    return _build_dp_job_response(job)



"""Run API router"""

from fastapi import APIRouter, status

from axion_lab_server.apps.api.deps import BatchPath, RunPath, RunRepo
from axion_lab_server.shared.domain import (
    CursorPaginatedResponse,
    RunCreate,
    RunResponse,
    RunStatus,
    RunSummaryResponse,
    RunUpdate,
)
from axion_lab_server.shared.domain.run import (
    OthersSection,
    RecentCollapsed,
    RunBriefResponse,
)

router = APIRouter(tags=["Runs"])


@router.post(
    "/batches/{batch_id}/runs",
    response_model=RunResponse,
    status_code=status.HTTP_201_CREATED,
)
async def create_run(
    data: RunCreate,
    batch: BatchPath,
    repo: RunRepo,
) -> RunResponse:
    """Create a new run in a batch"""
    run = await repo.create(batch.batch_id, data)
    return RunResponse.model_validate(run)


@router.get(
    "/batches/{batch_id}/runs",
    response_model=CursorPaginatedResponse[RunResponse],
)
async def list_runs(
    batch: BatchPath,
    repo: RunRepo,
    status: RunStatus | None = None,
    include_garbage: bool = False,
    tag: str | None = None,
    q: str | None = None,
    limit: int = 20,
    cursor: str | None = None,
) -> CursorPaginatedResponse[RunResponse]:
    """List runs in a batch with filtering"""
    runs, next_cursor = await repo.list_by_batch(
        batch.batch_id,
        status=status,
        include_garbage=include_garbage,
        tag=tag,
        q=q,
        limit=limit,
        cursor=cursor,
    )
    return CursorPaginatedResponse(
        items=[RunResponse.model_validate(r) for r in runs],
        next_cursor=next_cursor,
        has_more=next_cursor is not None,
    )


@router.get(
    "/batches/{batch_id}/runs/summary",
    response_model=RunSummaryResponse,
)
async def get_runs_summary(
    batch: BatchPath,
    repo: RunRepo,
    include_garbage: bool = False,
) -> RunSummaryResponse:
    """Get run summary with champion, recent, user_selected, and others"""
    # Get champion
    champion = await repo.get_champion(batch.batch_id)

    # Get user selected
    user_selected = await repo.get_user_selected(batch.batch_id)

    # Collect IDs to exclude from recent and others
    exclude_ids: list[str] = []
    if champion:
        exclude_ids.append(champion.run_id)
    exclude_ids.extend([r.run_id for r in user_selected])

    # Get recent 3
    recent = await repo.get_recent(batch.batch_id, limit=3, exclude_ids=exclude_ids)
    exclude_ids.extend([r.run_id for r in recent])

    # Get others
    others, next_cursor = await repo.list_by_batch(
        batch.batch_id,
        include_garbage=include_garbage,
        limit=10,
    )
    # Filter out already included runs
    others = [r for r in others if r.run_id not in exclude_ids]

    return RunSummaryResponse(
        champion=RunBriefResponse.model_validate(champion) if champion else None,
        recentCollapsed=RecentCollapsed(
            default_open=False,
            runs=[RunBriefResponse.model_validate(r) for r in recent],
        ),
        userSelected=[RunBriefResponse.model_validate(r) for r in user_selected],
        others=OthersSection(
            cursor=next_cursor,
            runs=[RunBriefResponse.model_validate(r) for r in others],
        ),
    )


@router.get("/runs/{run_id}", response_model=RunResponse)
async def get_run(run: RunPath) -> RunResponse:
    """Get run by ID"""
    return RunResponse.model_validate(run)


@router.patch("/runs/{run_id}", response_model=RunResponse)
async def update_run(
    data: RunUpdate,
    run: RunPath,
    repo: RunRepo,
) -> RunResponse:
    """Update a run"""
    updated = await repo.update(run.run_id, data)
    if not updated:
        # This shouldn't happen since we already validated the run exists
        raise RuntimeError("Run update failed unexpectedly")
    return RunResponse.model_validate(updated)



from datetime import UTC, datetime
from types import SimpleNamespace
from unittest.mock import AsyncMock, Mock

import pytest

from axion_lab_server.apps.api.routers.orgs import create_org, get_org, list_orgs
from axion_lab_server.shared.domain import OrgCreate


def _org(*, org_id: str = "org-1", name: str = "Acme Corp"):
    return SimpleNamespace(
        org_id=org_id,
        name=name,
        created_at=datetime(2026, 2, 12, tzinfo=UTC),
    )


@pytest.mark.asyncio
async def test_create_org_returns_response() -> None:
    repo = Mock()
    org = _org()
    repo.create = AsyncMock(return_value=org)

    data = OrgCreate(name="Acme Corp")
    response = await create_org(data=data, repo=repo)

    repo.create.assert_awaited_once_with(data)
    assert response.org_id == "org-1"
    assert response.name == "Acme Corp"


@pytest.mark.asyncio
async def test_list_orgs_returns_cursor_page() -> None:
    repo = Mock()
    orgs = [_org(org_id="org-1"), _org(org_id="org-2")]
    repo.list_all = AsyncMock(return_value=(orgs, "next-cursor"))

    page = await list_orgs(repo=repo, limit=2, cursor="cur-1")

    repo.list_all.assert_awaited_once_with(limit=2, cursor="cur-1")
    assert page.has_more is True
    assert page.next_cursor == "next-cursor"
    assert [item.org_id for item in page.items] == ["org-1", "org-2"]


@pytest.mark.asyncio
async def test_list_orgs_has_more_false_without_cursor() -> None:
    repo = Mock()
    repo.list_all = AsyncMock(return_value=([_org()], None))

    page = await list_orgs(repo=repo)

    assert page.has_more is False
    assert page.next_cursor is None
    assert len(page.items) == 1


@pytest.mark.asyncio
async def test_get_org_returns_validated_response() -> None:
    org = _org(org_id="org-99", name="Test Org")

    response = await get_org(org=org)

    assert response.org_id == "org-99"
    assert response.name == "Test Org"



"""DP (Differential Processing) service."""

from __future__ import annotations

from axion_lab.schemas import DPJobCreate, DPJobMode, DPJobResponse
from axion_lab.services._http import HttpTransport


class DPService:
    """Operations for DP computation jobs."""

    def __init__(self, http: HttpTransport) -> None:
        self._http = http

    def trigger(
        self,
        batch_id: str,
        *,
        mode: DPJobMode | str = DPJobMode.ACTIVE_ONLY,
        recompute: bool = False,
    ) -> DPJobResponse:
        """Trigger a DP computation job for a batch."""
        body = DPJobCreate(mode=DPJobMode(mode), recompute=recompute)
        data = self._http.post(
            f"/batches/{batch_id}/dp/compute",
            json=body.model_dump(),
        )
        return DPJobResponse.model_validate(data)

    def get_job(self, job_id: str) -> DPJobResponse:
        """Get the status of a DP job."""
        data = self._http.get(f"/dp/jobs/{job_id}")
        return DPJobResponse.model_validate(data)



"""Aggregation API router"""

import json

from axion_lab_server.apps.api.deps import (
    AggregationPath,
    AggregationRepo,
    CIRepo,
    ProjectPath,
    QMRepo,
)
from axion_lab_server.shared.domain import (
    AggregationCreate,
    AggregationMemberAdd,
    AggregationMemberResponse,
    AggregationResponse,
    ComparisonIndicatorResponse,
    CursorPaginatedResponse,
    QualityMetricResponse,
)
from fastapi import APIRouter, status

router = APIRouter(tags=["Aggregations"])


def _build_aggregation_response(agg, *, member_count: int = 0) -> AggregationResponse:
    """Build aggregation response from ORM entity"""
    return AggregationResponse(
        aggregationId=agg.aggregation_id,
        projectId=agg.project_id,
        name=agg.name,
        description=agg.description,
        groupByKeys=json.loads(agg.group_by_keys_json),
        filter=json.loads(agg.filter_json),
        createdAt=agg.created_at,
        memberCount=member_count,
    )


def _build_member_response(member) -> AggregationMemberResponse:
    """Build aggregation member response from ORM entity"""
    return AggregationMemberResponse(
        memberId=member.member_id,
        aggregationId=member.aggregation_id,
        runId=member.run_id,
        metadata=json.loads(member.metadata_json),
        addedAt=member.added_at,
    )


# ── Aggregation CRUD ──────────────────────────────────────────


@router.post(
    "/projects/{project_id}/aggregations",
    response_model=AggregationResponse,
    status_code=status.HTTP_201_CREATED,
)
async def create_aggregation(
    data: AggregationCreate,
    project: ProjectPath,
    repo: AggregationRepo,
) -> AggregationResponse:
    """Create a new aggregation in a project"""
    agg = await repo.create(project.project_id, data)
    return _build_aggregation_response(agg, member_count=0)


@router.get(
    "/projects/{project_id}/aggregations",
    response_model=CursorPaginatedResponse[AggregationResponse],
)
async def list_aggregations(
    project: ProjectPath,
    repo: AggregationRepo,
    limit: int = 20,
    cursor: str | None = None,
) -> CursorPaginatedResponse[AggregationResponse]:
    """List aggregations in a project"""
    aggregations, next_cursor = await repo.list_by_project(
        project.project_id, limit=limit, cursor=cursor
    )
    items = []
    for agg in aggregations:
        count = await repo.member_count(agg.aggregation_id)
        items.append(_build_aggregation_response(agg, member_count=count))
    return CursorPaginatedResponse(
        items=items,
        next_cursor=next_cursor,
        has_more=next_cursor is not None,
    )


@router.get(
    "/aggregations/{aggregation_id}",
    response_model=AggregationResponse,
)
async def get_aggregation(
    aggregation: AggregationPath,
    repo: AggregationRepo,
) -> AggregationResponse:
    """Get aggregation by ID"""
    count = await repo.member_count(aggregation.aggregation_id)
    return _build_aggregation_response(aggregation, member_count=count)


@router.delete(
    "/aggregations/{aggregation_id}",
    status_code=status.HTTP_204_NO_CONTENT,
)
async def delete_aggregation(
    aggregation: AggregationPath,
    repo: AggregationRepo,
) -> None:
    """Delete an aggregation"""
    await repo.delete(aggregation.aggregation_id)


# ── Member management ─────────────────────────────────────────


@router.post(
    "/aggregations/{aggregation_id}/members",
    response_model=AggregationMemberResponse,
    status_code=status.HTTP_201_CREATED,
)
async def add_member(
    data: AggregationMemberAdd,
    aggregation: AggregationPath,
    repo: AggregationRepo,
) -> AggregationMemberResponse:
    """Add a run to an aggregation"""
    member = await repo.add_member(aggregation.aggregation_id, data)
    return _build_member_response(member)


@router.get(
    "/aggregations/{aggregation_id}/members",
    response_model=CursorPaginatedResponse[AggregationMemberResponse],
)
async def list_members(
    aggregation: AggregationPath,
    repo: AggregationRepo,
    limit: int = 100,
    cursor: str | None = None,
) -> CursorPaginatedResponse[AggregationMemberResponse]:
    """List members of an aggregation"""
    members, next_cursor = await repo.list_members(
        aggregation.aggregation_id, limit=limit, cursor=cursor
    )
    return CursorPaginatedResponse(
        items=[_build_member_response(m) for m in members],
        next_cursor=next_cursor,
        has_more=next_cursor is not None,
    )


@router.delete(
    "/aggregations/{aggregation_id}/members/{run_id}",
    status_code=status.HTTP_204_NO_CONTENT,
)
async def remove_member(
    aggregation: AggregationPath,
    run_id: str,
    repo: AggregationRepo,
) -> None:
    """Remove a run from an aggregation"""
    await repo.remove_member(aggregation.aggregation_id, run_id)


# ── Metrics across aggregation members ────────────────────────


def _build_qm_response(qm) -> QualityMetricResponse:
    """Build quality metric response"""
    value = json.loads(qm.value_json) if qm.value_json else None
    return QualityMetricResponse(
        qmId=qm.qm_id,
        runId=qm.run_id,
        key=qm.key,
        value=value,
        source=qm.source,
        computedAt=qm.computed_at,
        version=qm.version,
    )


def _build_ci_response(ci) -> ComparisonIndicatorResponse:
    """Build comparison indicator response"""
    value = json.loads(ci.value_json) if ci.value_json else None
    return ComparisonIndicatorResponse(
        ciId=ci.ci_id,
        runId=ci.run_id,
        key=ci.key,
        value=value,
        baselineRef=ci.baseline_ref,
        computedAt=ci.computed_at,
        version=ci.version,
    )


@router.get(
    "/aggregations/{aggregation_id}/quality-metrics",
    response_model=CursorPaginatedResponse[QualityMetricResponse],
)
async def list_quality_metrics_by_aggregation(
    aggregation: AggregationPath,
    repo: QMRepo,
    key: str | None = None,
    limit: int = 100,
    cursor: str | None = None,
) -> CursorPaginatedResponse[QualityMetricResponse]:
    """List quality metrics for all runs in an aggregation"""
    metrics, next_cursor = await repo.list_by_aggregation(
        aggregation.aggregation_id, key=key, limit=limit, cursor=cursor
    )
    return CursorPaginatedResponse(
        items=[_build_qm_response(m) for m in metrics],
        next_cursor=next_cursor,
        has_more=next_cursor is not None,
    )


@router.get(
    "/aggregations/{aggregation_id}/comparison-indicators",
    response_model=CursorPaginatedResponse[ComparisonIndicatorResponse],
)
async def list_comparison_indicators_by_aggregation(
    aggregation: AggregationPath,
    repo: CIRepo,
    key: str | None = None,
    limit: int = 100,
    cursor: str | None = None,
) -> CursorPaginatedResponse[ComparisonIndicatorResponse]:
    """List comparison indicators for all runs in an aggregation"""
    indicators, next_cursor = await repo.list_by_aggregation(
        aggregation.aggregation_id, key=key, limit=limit, cursor=cursor
    )
    return CursorPaginatedResponse(
        items=[_build_ci_response(ci) for ci in indicators],
        next_cursor=next_cursor,
        has_more=next_cursor is not None,
    )



"""S3/MinIO object storage implementation"""

import json
from contextlib import asynccontextmanager
from typing import Any

import aioboto3

from axion_lab_server.gateways.storage.base import ObjectRef, ObjectStore


class S3ObjectStore(ObjectStore):
    """S3/MinIO implementation of ObjectStore"""

    def __init__(
        self,
        bucket: str,
        endpoint_url: str | None = None,
        access_key: str | None = None,
        secret_key: str | None = None,
        region: str = "us-east-1",
    ):
        self.bucket = bucket
        self.endpoint_url = endpoint_url
        self.access_key = access_key
        self.secret_key = secret_key
        self.region = region
        self._session = aioboto3.Session()

    @asynccontextmanager
    async def _get_client(self):
        """Get S3 client as context manager"""
        async with self._session.client(
            "s3",
            endpoint_url=self.endpoint_url,
            aws_access_key_id=self.access_key,
            aws_secret_access_key=self.secret_key,
            region_name=self.region,
        ) as client:
            yield client

    async def put_bytes(
        self,
        key: str,
        data: bytes,
        content_type: str | None = None,
        metadata: dict[str, str] | None = None,
    ) -> ObjectRef:
        """Store bytes at the given key"""
        extra_args: dict[str, Any] = {}
        if content_type:
            extra_args["ContentType"] = content_type
        if metadata:
            extra_args["Metadata"] = metadata

        async with self._get_client() as client:
            await client.put_object(
                Bucket=self.bucket,
                Key=key,
                Body=data,
                **extra_args,
            )

        provider = "minio" if self.endpoint_url else "s3"
        return ObjectRef(
            key=key,
            bucket=self.bucket,
            provider=provider,
            size=len(data),
            content_type=content_type,
            metadata=metadata,
        )

    async def put_json(
        self,
        key: str,
        data: Any,
        metadata: dict[str, str] | None = None,
    ) -> ObjectRef:
        """Store JSON data at the given key"""
        json_bytes = json.dumps(data, ensure_ascii=False, indent=2).encode("utf-8")
        return await self.put_bytes(
            key, json_bytes, content_type="application/json", metadata=metadata
        )

    async def get_bytes(self, key: str) -> bytes:
        """Retrieve bytes from the given key"""
        async with self._get_client() as client:
            response = await client.get_object(Bucket=self.bucket, Key=key)
            async with response["Body"] as stream:
                return await stream.read()

    async def get_json(self, key: str) -> Any:
        """Retrieve JSON data from the given key"""
        data = await self.get_bytes(key)
        return json.loads(data.decode("utf-8"))

    async def exists(self, key: str) -> bool:
        """Check if an object exists at the given key"""
        try:
            async with self._get_client() as client:
                await client.head_object(Bucket=self.bucket, Key=key)
            return True
        except Exception:
            return False

    async def list_keys(self, prefix: str) -> list[ObjectRef]:
        """List objects with the given prefix"""
        results: list[ObjectRef] = []
        provider = "minio" if self.endpoint_url else "s3"

        async with self._get_client() as client:
            paginator = client.get_paginator("list_objects_v2")
            async for page in paginator.paginate(Bucket=self.bucket, Prefix=prefix):
                for obj in page.get("Contents", []):
                    results.append(
                        ObjectRef(
                            key=obj["Key"],
                            bucket=self.bucket,
                            provider=provider,
                            size=obj.get("Size"),
                        )
                    )

        return results

    async def delete(self, key: str) -> bool:
        """Delete the object at the given key"""
        try:
            async with self._get_client() as client:
                await client.delete_object(Bucket=self.bucket, Key=key)
            return True
        except Exception:
            return False

    async def presign_get(self, key: str, expires_sec: int = 3600) -> str:
        """Generate a presigned URL for GET operations"""
        async with self._get_client() as client:
            url = await client.generate_presigned_url(
                "get_object",
                Params={"Bucket": self.bucket, "Key": key},
                ExpiresIn=expires_sec,
            )
        return url

    async def presign_put(
        self, key: str, content_type: str | None = None, expires_sec: int = 3600
    ) -> str:
        """Generate a presigned URL for PUT operations"""
        params: dict[str, Any] = {"Bucket": self.bucket, "Key": key}
        if content_type:
            params["ContentType"] = content_type

        async with self._get_client() as client:
            url = await client.generate_presigned_url(
                "put_object",
                Params=params,
                ExpiresIn=expires_sec,
            )
        return url



"""Artifact schemas"""

import json
from datetime import datetime
from enum import StrEnum
from typing import Any

from pydantic import BaseModel, Field, field_validator, model_validator


class ArtifactKind(StrEnum):
    """Artifact kind enumeration"""

    URL = "url"
    LOCAL = "local"
    INLINE_TEXT = "inline_text"
    INLINE_NUMBER = "inline_number"
    INLINE_JSON = "inline_json"


class ArtifactType(StrEnum):
    """Artifact type enumeration"""

    LANGFUSE_TRACE = "langfuse_trace"
    OBJECT_STORAGE = "object_storage"
    NOTE = "note"
    DASHBOARD = "dashboard"
    GIT_COMMIT = "git_commit"
    FILE = "file"
    EVALUATION = "evaluation"
    LATENCY_P95_MS = "latency_p95_ms"
    COST_USD = "cost_usd"
    OTHER = "other"


class ArtifactCreate(BaseModel):
    """Schema for creating an artifact"""

    kind: ArtifactKind
    type: str = Field(..., min_length=1, max_length=50)
    label: str = Field(..., min_length=1, max_length=255)
    payload: Any  # URL/path string, number, or JSON object
    meta: dict[str, Any] = Field(default_factory=dict)

    @model_validator(mode="after")
    def validate_payload_for_kind(self) -> "ArtifactCreate":
        """Validate payload matches kind"""
        if self.kind in (
            ArtifactKind.URL,
            ArtifactKind.LOCAL,
            ArtifactKind.INLINE_TEXT,
        ):
            if not isinstance(self.payload, str):
                raise ValueError(f"payload must be string for kind={self.kind}")
        elif self.kind == ArtifactKind.INLINE_NUMBER:
            if not isinstance(self.payload, int | float):
                raise ValueError(f"payload must be number for kind={self.kind}")
        # INLINE_JSON can be any JSON-serializable value
        return self


class ArtifactResponse(BaseModel):
    """Schema for artifact response"""

    artifact_id: str = Field(..., alias="artifactId")
    run_id: str = Field(..., alias="runId")
    kind: ArtifactKind
    type: str
    label: str
    payload: Any
    meta: dict[str, Any] = Field(default_factory=dict)
    created_at: datetime = Field(..., alias="createdAt")

    model_config = {"populate_by_name": True, "from_attributes": True}

    @field_validator("meta", mode="before")
    @classmethod
    def parse_meta(cls, v: Any) -> dict[str, Any]:
        """Parse meta from JSON string if needed"""
        if isinstance(v, str):
            return json.loads(v)
        return v



"""Run schemas"""

import json
from datetime import datetime
from enum import StrEnum
from typing import Any

from pydantic import BaseModel, Field, field_validator


class RunStatus(StrEnum):
    """Run status enumeration"""

    ACTIVE = "active"
    GARBAGE = "garbage"
    ARCHIVED = "archived"


class RunCreate(BaseModel):
    """Schema for creating a run"""

    name: str = Field(..., min_length=1, max_length=255)
    status: RunStatus = RunStatus.ACTIVE
    tags: list[str] = Field(default_factory=list)
    note: str | None = None


class RunUpdate(BaseModel):
    """Schema for updating a run"""

    name: str | None = Field(None, min_length=1, max_length=255)
    status: RunStatus | None = None
    tags: list[str] | None = None
    note: str | None = None


class RunResponse(BaseModel):
    """Schema for run response"""

    run_id: str = Field(..., alias="runId")
    batch_id: str = Field(..., alias="batchId")
    name: str
    status: RunStatus
    tags: list[str] = Field(default_factory=list)
    note: str | None = None
    created_at: datetime = Field(..., alias="createdAt")
    updated_at: datetime = Field(..., alias="updatedAt")

    model_config = {"populate_by_name": True, "from_attributes": True}

    @field_validator("tags", mode="before")
    @classmethod
    def parse_tags(cls, v: Any) -> list[str]:
        """Parse tags from JSON string if needed"""
        if isinstance(v, str):
            return json.loads(v)
        return v


class RunBriefResponse(BaseModel):
    """Brief run response for summary views"""

    run_id: str = Field(..., alias="runId")
    name: str

    model_config = {"populate_by_name": True, "from_attributes": True}


class RecentCollapsed(BaseModel):
    """Recent runs collapsed section"""

    default_open: bool = Field(default=False, alias="defaultOpen")
    runs: list[RunBriefResponse]

    model_config = {"populate_by_name": True}


class OthersSection(BaseModel):
    """Others section with cursor pagination"""

    cursor: str | None = None
    runs: list[RunBriefResponse]


class RunSummaryResponse(BaseModel):
    """Run summary response with champion, recent, user_selected, and others"""

    champion: RunBriefResponse | None = None
    recent_collapsed: RecentCollapsed = Field(..., alias="recentCollapsed")
    user_selected: list[RunBriefResponse] = Field(
        default_factory=list, alias="userSelected"
    )
    others: OthersSection

    model_config = {"populate_by_name": True}



"""Aggregation repository"""

import json

from axion_lab_server.repos.models.entities import Aggregation, AggregationMember
from axion_lab_server.shared.domain import AggregationCreate, AggregationMemberAdd
from axion_lab_server.shared.libs import generate_id, utc_now
from sqlalchemy import delete, func, select
from sqlalchemy.ext.asyncio import AsyncSession


class AggregationRepository:
    """Repository for Aggregation and AggregationMember operations"""

    def __init__(self, session: AsyncSession):
        self.session = session

    # ── Aggregation CRUD ──────────────────────────────────────

    async def create(self, project_id: str, data: AggregationCreate) -> Aggregation:
        """Create a new aggregation"""
        aggregation = Aggregation(
            aggregation_id=generate_id(),
            project_id=project_id,
            name=data.name,
            description=data.description,
            group_by_keys_json=json.dumps(data.group_by_keys),
            filter_json=json.dumps(data.filter),
            created_at=utc_now(),
        )
        self.session.add(aggregation)
        await self.session.flush()
        return aggregation

    async def get_by_id(self, aggregation_id: str) -> Aggregation | None:
        """Get aggregation by ID"""
        result = await self.session.execute(
            select(Aggregation).where(Aggregation.aggregation_id == aggregation_id)
        )
        return result.scalar_one_or_none()

    async def list_by_project(
        self,
        project_id: str,
        limit: int = 20,
        cursor: str | None = None,
    ) -> tuple[list[Aggregation], str | None]:
        """List aggregations by project with cursor pagination"""
        query = (
            select(Aggregation)
            .where(Aggregation.project_id == project_id)
            .order_by(Aggregation.created_at.desc())
        )

        if cursor:
            cursor_agg = await self.get_by_id(cursor)
            if cursor_agg:
                query = query.where(Aggregation.created_at < cursor_agg.created_at)

        query = query.limit(limit + 1)
        result = await self.session.execute(query)
        aggregations = list(result.scalars().all())

        next_cursor = None
        if len(aggregations) > limit:
            aggregations = aggregations[:limit]
            next_cursor = aggregations[-1].aggregation_id

        return aggregations, next_cursor

    async def delete(self, aggregation_id: str) -> bool:
        """Delete an aggregation and its members"""
        agg = await self.get_by_id(aggregation_id)
        if not agg:
            return False
        await self.session.delete(agg)
        await self.session.flush()
        return True

    async def member_count(self, aggregation_id: str) -> int:
        """Count members of an aggregation"""
        result = await self.session.execute(
            select(func.count()).where(
                AggregationMember.aggregation_id == aggregation_id
            )
        )
        return result.scalar_one()

    # ── Member CRUD ───────────────────────────────────────────

    async def add_member(
        self, aggregation_id: str, data: AggregationMemberAdd
    ) -> AggregationMember:
        """Add a run to an aggregation"""
        member = AggregationMember(
            member_id=generate_id(),
            aggregation_id=aggregation_id,
            run_id=data.run_id,
            metadata_json=json.dumps(data.metadata),
            added_at=utc_now(),
        )
        self.session.add(member)
        await self.session.flush()
        return member

    async def remove_member(self, aggregation_id: str, run_id: str) -> bool:
        """Remove a run from an aggregation"""
        result = await self.session.execute(
            delete(AggregationMember).where(
                AggregationMember.aggregation_id == aggregation_id,
                AggregationMember.run_id == run_id,
            )
        )
        rowcount = getattr(result, "rowcount", 0)
        return bool(rowcount)

    async def list_members(
        self,
        aggregation_id: str,
        limit: int = 100,
        cursor: str | None = None,
    ) -> tuple[list[AggregationMember], str | None]:
        """List members of an aggregation with cursor pagination"""
        query = (
            select(AggregationMember)
            .where(AggregationMember.aggregation_id == aggregation_id)
            .order_by(AggregationMember.added_at.desc())
        )

        if cursor:
            cursor_member = await self.session.execute(
                select(AggregationMember).where(AggregationMember.member_id == cursor)
            )
            cursor_obj = cursor_member.scalar_one_or_none()
            if cursor_obj:
                query = query.where(AggregationMember.added_at < cursor_obj.added_at)

        query = query.limit(limit + 1)
        result = await self.session.execute(query)
        members = list(result.scalars().all())

        next_cursor = None
        if len(members) > limit:
            members = members[:limit]
            next_cursor = members[-1].member_id

        return members, next_cursor

    async def get_member_run_ids(self, aggregation_id: str) -> list[str]:
        """Get all run IDs in an aggregation (for metric queries)"""
        result = await self.session.execute(
            select(AggregationMember.run_id).where(
                AggregationMember.aggregation_id == aggregation_id
            )
        )
        return list(result.scalars().all())



"""Comparison Indicator service."""

from __future__ import annotations

from axion_lab.schemas import ComparisonIndicatorResponse, CursorPaginatedResponse
from axion_lab.services._http import HttpTransport


class ComparisonIndicatorService:
    """Read-only access to comparison indicators (computed via DP)."""

    def __init__(self, http: HttpTransport) -> None:
        self._http = http

    def list_by_run(
        self,
        run_id: str,
        *,
        limit: int = 100,
        cursor: str | None = None,
    ) -> CursorPaginatedResponse[ComparisonIndicatorResponse]:
        """List comparison indicators for a run."""
        params: dict[str, object] = {"limit": limit}
        if cursor:
            params["cursor"] = cursor
        data = self._http.get(f"/runs/{run_id}/comparison-indicators", params=params)
        return CursorPaginatedResponse[ComparisonIndicatorResponse].model_validate(data)

    def list_by_batch(
        self,
        batch_id: str,
        *,
        key: str | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> CursorPaginatedResponse[ComparisonIndicatorResponse]:
        """List comparison indicators for all runs in a batch."""
        params: dict[str, object] = {"limit": limit}
        if key:
            params["key"] = key
        if cursor:
            params["cursor"] = cursor
        data = self._http.get(
            f"/batches/{batch_id}/comparison-indicators", params=params
        )
        return CursorPaginatedResponse[ComparisonIndicatorResponse].model_validate(data)



from datetime import UTC, datetime
from types import SimpleNamespace
from unittest.mock import AsyncMock, Mock

import pytest

from axion_lab_server.apps.api.routers.projects import (
    create_project,
    get_project,
    list_projects,
)
from axion_lab_server.shared.domain import ProjectCreate


def _org(*, org_id: str = "org-1"):
    return SimpleNamespace(org_id=org_id)


def _project(
    *, project_id: str = "proj-1", org_id: str = "org-1", name: str = "My Project"
):
    return SimpleNamespace(
        project_id=project_id,
        org_id=org_id,
        name=name,
        created_at=datetime(2026, 2, 12, tzinfo=UTC),
    )


@pytest.mark.asyncio
async def test_create_project_returns_response() -> None:
    org = _org()
    repo = Mock()
    project = _project()
    repo.create = AsyncMock(return_value=project)

    data = ProjectCreate(name="My Project")
    response = await create_project(data=data, org=org, repo=repo)

    repo.create.assert_awaited_once_with("org-1", data)
    assert response.project_id == "proj-1"
    assert response.org_id == "org-1"
    assert response.name == "My Project"


@pytest.mark.asyncio
async def test_list_projects_returns_cursor_page() -> None:
    org = _org()
    repo = Mock()
    projects = [_project(project_id="proj-1"), _project(project_id="proj-2")]
    repo.list_by_org = AsyncMock(return_value=(projects, "next-cursor"))

    page = await list_projects(org=org, repo=repo, limit=2, cursor="cur-1")

    repo.list_by_org.assert_awaited_once_with("org-1", limit=2, cursor="cur-1")
    assert page.has_more is True
    assert page.next_cursor == "next-cursor"
    assert [item.project_id for item in page.items] == ["proj-1", "proj-2"]


@pytest.mark.asyncio
async def test_list_projects_has_more_false_without_cursor() -> None:
    org = _org()
    repo = Mock()
    repo.list_by_org = AsyncMock(return_value=([_project()], None))

    page = await list_projects(org=org, repo=repo)

    assert page.has_more is False
    assert page.next_cursor is None
    assert len(page.items) == 1


@pytest.mark.asyncio
async def test_get_project_returns_validated_response() -> None:
    project = _project(project_id="proj-99", name="Test Project")

    response = await get_project(project=project)

    assert response.project_id == "proj-99"
    assert response.name == "Test Project"



from datetime import UTC, datetime

from axion_lab_server.shared.libs.utils import generate_id, utc_now


def test_generate_id_returns_string() -> None:
    id_ = generate_id()

    assert isinstance(id_, str)
    assert len(id_) > 0


def test_generate_id_returns_unique_values() -> None:
    ids = {generate_id() for _ in range(100)}

    assert len(ids) == 100


def test_utc_now_returns_utc_datetime() -> None:
    now = utc_now()

    assert isinstance(now, datetime)
    assert now.tzinfo == UTC


def test_utc_now_is_recent() -> None:
    before = datetime.now(UTC)
    now = utc_now()
    after = datetime.now(UTC)

    assert before <= now <= after



"""Artifact service."""

from __future__ import annotations

from typing import Any

from axion_lab.schemas import (
    ArtifactCreate,
    ArtifactKind,
    ArtifactResponse,
    CursorPaginatedResponse,
)
from axion_lab.services._http import HttpTransport


class ArtifactService:
    """CRUD operations for artifacts."""

    def __init__(self, http: HttpTransport) -> None:
        self._http = http

    def create(
        self,
        run_id: str,
        *,
        kind: ArtifactKind | str,
        type: str,
        label: str,
        payload: Any,
        meta: dict[str, Any] | None = None,
    ) -> ArtifactResponse:
        """Create a new artifact for a run."""
        body = ArtifactCreate(
            kind=ArtifactKind(kind),
            type=type,
            label=label,
            payload=payload,
            meta=meta or {},
        )
        data = self._http.post(
            f"/runs/{run_id}/artifacts",
            json=body.model_dump(),
        )
        return ArtifactResponse.model_validate(data)

    def list(
        self,
        run_id: str,
        *,
        limit: int = 100,
        cursor: str | None = None,
    ) -> CursorPaginatedResponse[ArtifactResponse]:
        """List artifacts for a run."""
        params: dict[str, object] = {"limit": limit}
        if cursor:
            params["cursor"] = cursor
        data = self._http.get(f"/runs/{run_id}/artifacts", params=params)
        return CursorPaginatedResponse[ArtifactResponse].model_validate(data)

    def delete(self, artifact_id: str) -> None:
        """Delete an artifact by ID."""
        self._http.delete(f"/artifacts/{artifact_id}")



"""Artifact repository"""

import json
from typing import Any

from sqlalchemy import delete, select
from sqlalchemy.ext.asyncio import AsyncSession

from axion_lab_server.repos.models.entities import Artifact
from axion_lab_server.shared.domain import ArtifactCreate, ArtifactKind
from axion_lab_server.shared.libs import generate_id, utc_now


class ArtifactRepository:
    """Repository for Artifact operations"""

    def __init__(self, session: AsyncSession):
        self.session = session

    async def create(self, run_id: str, data: ArtifactCreate) -> Artifact:
        """Create a new artifact"""
        payload_text: str | None = None
        payload_number: float | None = None

        if data.kind == ArtifactKind.INLINE_NUMBER:
            payload_number = float(data.payload)
        elif data.kind == ArtifactKind.INLINE_JSON:
            payload_text = json.dumps(data.payload)
        else:
            payload_text = str(data.payload)

        artifact = Artifact(
            artifact_id=generate_id(),
            run_id=run_id,
            kind=data.kind.value,
            type=data.type,
            label=data.label,
            payload_text=payload_text,
            payload_number=payload_number,
            meta_json=json.dumps(data.meta),
            created_at=utc_now(),
        )
        self.session.add(artifact)
        await self.session.flush()
        return artifact

    async def get_by_id(self, artifact_id: str) -> Artifact | None:
        """Get artifact by ID"""
        result = await self.session.execute(
            select(Artifact).where(Artifact.artifact_id == artifact_id)
        )
        return result.scalar_one_or_none()

    async def list_by_run(
        self, run_id: str, limit: int = 100, cursor: str | None = None
    ) -> tuple[list[Artifact], str | None]:
        """List artifacts by run with cursor pagination"""
        query = (
            select(Artifact)
            .where(Artifact.run_id == run_id)
            .order_by(Artifact.created_at.desc())
        )

        if cursor:
            cursor_artifact = await self.get_by_id(cursor)
            if cursor_artifact:
                query = query.where(Artifact.created_at < cursor_artifact.created_at)

        query = query.limit(limit + 1)
        result = await self.session.execute(query)
        artifacts = list(result.scalars().all())

        next_cursor = None
        if len(artifacts) > limit:
            artifacts = artifacts[:limit]
            next_cursor = artifacts[-1].artifact_id

        return artifacts, next_cursor

    async def list_by_run_and_type(
        self, run_id: str, artifact_type: str
    ) -> list[Artifact]:
        """List artifacts by run and type"""
        result = await self.session.execute(
            select(Artifact)
            .where(Artifact.run_id == run_id, Artifact.type == artifact_type)
            .order_by(Artifact.created_at.desc())
        )
        return list(result.scalars().all())

    async def delete(self, artifact_id: str) -> bool:
        """Delete an artifact"""
        result = await self.session.execute(
            delete(Artifact).where(Artifact.artifact_id == artifact_id)
        )
        rowcount = getattr(result, "rowcount", 0)
        return int(rowcount or 0) > 0

    def get_payload(self, artifact: Artifact) -> Any:
        """Get artifact payload in appropriate type"""
        if artifact.kind == ArtifactKind.INLINE_NUMBER.value:
            return artifact.payload_number
        elif artifact.kind == ArtifactKind.INLINE_JSON.value:
            return json.loads(artifact.payload_text) if artifact.payload_text else None
        else:
            return artifact.payload_text



from __future__ import annotations

from dataclasses import dataclass
from types import SimpleNamespace

import pytest

from axion_lab.high_level import create_artifact, create_run, set_active_run


@dataclass
class _Named:
    name: str
    id_value: str

    @property
    def org_id(self) -> str:
        return self.id_value

    @property
    def project_id(self) -> str:
        return self.id_value

    @property
    def batch_id(self) -> str:
        return self.id_value


class _FakeOrgService:
    def __init__(self, items: list[_Named]) -> None:
        self._items = items

    def list(self, *, limit: int = 100, cursor: str | None = None):
        return SimpleNamespace(items=self._items, next_cursor=None)

    def create(self, name: str):
        item = _Named(name=name, id_value="org-created")
        self._items.append(item)
        return item


class _FakeProjectService:
    def __init__(self, items: list[_Named]) -> None:
        self._items = items

    def list(self, org_id: str, *, limit: int = 100, cursor: str | None = None):
        return SimpleNamespace(items=self._items, next_cursor=None)

    def create(self, org_id: str, name: str):
        item = _Named(name=name, id_value="project-created")
        self._items.append(item)
        return item


class _FakeBatchService:
    def __init__(self, items: list[_Named]) -> None:
        self._items = items

    def list(self, project_id: str, *, limit: int = 100, cursor: str | None = None):
        return SimpleNamespace(items=self._items, next_cursor=None)

    def create(self, project_id: str, name: str):
        item = _Named(name=name, id_value="batch-created")
        self._items.append(item)
        return item


class _FakeRunService:
    def __init__(self) -> None:
        self.calls: list[tuple[str, str, list[str] | None, str | None]] = []

    def create(
        self,
        batch_id: str,
        name: str,
        *,
        tags: list[str] | None = None,
        note: str | None = None,
    ):
        self.calls.append((batch_id, name, tags, note))
        return SimpleNamespace(run_id="run-001", name=name)


class _FakeArtifactService:
    def __init__(self) -> None:
        self.calls: list[dict[str, object]] = []

    def create(
        self,
        run_id: str,
        *,
        kind: str,
        type: str,
        label: str,
        payload: object,
        meta: dict[str, object] | None = None,
    ):
        call = {
            "run_id": run_id,
            "kind": kind,
            "type": type,
            "label": label,
            "payload": payload,
            "meta": meta,
        }
        self.calls.append(call)
        return SimpleNamespace(artifact_id="art-001", **call)


class _FakeClient:
    def __init__(
        self,
        org_items: list[_Named],
        project_items: list[_Named],
        batch_items: list[_Named],
    ) -> None:
        self.orgs = _FakeOrgService(org_items)
        self.projects = _FakeProjectService(project_items)
        self.batches = _FakeBatchService(batch_items)
        self.runs = _FakeRunService()
        self.artifacts = _FakeArtifactService()


@pytest.fixture(autouse=True)
def _reset_active_run() -> None:
    set_active_run(None)


@pytest.fixture
def _env_context(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("AXION_LAB_ORG", "Org-A")
    monkeypatch.setenv("AXION_LAB_PROJECT", "Project-A")
    monkeypatch.setenv("AXION_LAB_BATCH", "Batch-A")


def test_create_run_uses_existing_context(_env_context: None) -> None:
    client = _FakeClient(
        org_items=[_Named(name="Org-A", id_value="org-1")],
        project_items=[_Named(name="Project-A", id_value="project-1")],
        batch_items=[_Named(name="Batch-A", id_value="batch-1")],
    )

    run = create_run("run-a", tags=["t1"], note="memo", client=client)  # type: ignore[arg-type]

    assert run.run_id == "run-001"
    assert client.runs.calls == [("batch-1", "run-a", ["t1"], "memo")]


def test_create_run_get_or_create_context(_env_context: None) -> None:
    client = _FakeClient(org_items=[], project_items=[], batch_items=[])

    run = create_run("run-b", client=client)  # type: ignore[arg-type]

    assert run.run_id == "run-001"
    assert client.runs.calls == [("batch-created", "run-b", None, None)]


def test_create_artifact_uses_active_run(_env_context: None) -> None:
    client = _FakeClient(
        org_items=[_Named(name="Org-A", id_value="org-1")],
        project_items=[_Named(name="Project-A", id_value="project-1")],
        batch_items=[_Named(name="Batch-A", id_value="batch-1")],
    )
    create_run("run-c", client=client)  # type: ignore[arg-type]

    artifact = create_artifact(
        kind="inline_number",
        type="evaluation",
        label="score",
        payload=0.9,
        client=client,  # type: ignore[arg-type]
    )

    assert artifact.run_id == "run-001"
    assert client.artifacts.calls[-1]["run_id"] == "run-001"


def test_create_artifact_requires_run_id_without_active_run(_env_context: None) -> None:
    client = _FakeClient(org_items=[], project_items=[], batch_items=[])

    with pytest.raises(ValueError, match="run_id is required"):
        create_artifact(
            kind="inline_text",
            type="note",
            label="memo",
            payload="hello",
            client=client,  # type: ignore[arg-type]
        )


def test_create_run_requires_env(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.delenv("AXION_LAB_ORG", raising=False)
    monkeypatch.delenv("AXION_LAB_PROJECT", raising=False)
    monkeypatch.delenv("AXION_LAB_BATCH", raising=False)
    client = _FakeClient(org_items=[], project_items=[], batch_items=[])

    with pytest.raises(ValueError, match="Missing required environment variables"):
        create_run("run-z", client=client)  # type: ignore[arg-type]



"""Gateways package - external service integrations"""

from axion_lab_server.gateways.storage import ObjectRef, ObjectStore, get_object_store

__all__ = ["ObjectStore", "ObjectRef", "get_object_store"]



"""Batch repository"""

from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession

from axion_lab_server.repos.models.entities import Batch
from axion_lab_server.shared.domain import BatchCreate
from axion_lab_server.shared.libs import generate_id, utc_now


class BatchRepository:
    """Repository for Batch operations"""

    def __init__(self, session: AsyncSession):
        self.session = session

    async def create(self, project_id: str, data: BatchCreate) -> Batch:
        """Create a new batch"""
        batch = Batch(
            batch_id=generate_id(),
            project_id=project_id,
            name=data.name,
            created_at=utc_now(),
        )
        self.session.add(batch)
        await self.session.flush()
        return batch

    async def get_by_id(self, batch_id: str) -> Batch | None:
        """Get batch by ID"""
        result = await self.session.execute(
            select(Batch).where(Batch.batch_id == batch_id)
        )
        return result.scalar_one_or_none()

    async def list_by_project(
        self, project_id: str, limit: int = 20, cursor: str | None = None
    ) -> tuple[list[Batch], str | None]:
        """List batches by project with cursor pagination"""
        query = (
            select(Batch)
            .where(Batch.project_id == project_id)
            .order_by(Batch.created_at.desc())
        )

        if cursor:
            cursor_batch = await self.get_by_id(cursor)
            if cursor_batch:
                query = query.where(Batch.created_at < cursor_batch.created_at)

        query = query.limit(limit + 1)
        result = await self.session.execute(query)
        batches = list(result.scalars().all())

        next_cursor = None
        if len(batches) > limit:
            batches = batches[:limit]
            next_cursor = batches[-1].batch_id

        return batches, next_cursor



"""Quality Metric repository"""

import json
from typing import Any

from axion_lab_server.repos.models.entities import AggregationMember, QualityMetric, Run
from axion_lab_server.shared.domain import QualityMetricSource
from axion_lab_server.shared.libs import generate_id, utc_now
from sqlalchemy import and_, delete, select
from sqlalchemy.ext.asyncio import AsyncSession


class QualityMetricRepository:
    """Repository for QualityMetric operations"""

    def __init__(self, session: AsyncSession):
        self.session = session

    async def upsert(
        self,
        run_id: str,
        key: str,
        value: Any,
        source: QualityMetricSource,
        version: int = 1,
    ) -> QualityMetric:
        """Upsert a quality metric (insert or update)"""
        # Check if exists
        existing = await self.get_by_run_and_key(run_id, key, version)

        if existing:
            existing.value_json = json.dumps(value)
            existing.source = source.value
            existing.computed_at = utc_now()
            await self.session.flush()
            return existing

        qm = QualityMetric(
            qm_id=generate_id(),
            run_id=run_id,
            key=key,
            value_json=json.dumps(value),
            source=source.value,
            computed_at=utc_now(),
            version=version,
        )
        self.session.add(qm)
        await self.session.flush()
        return qm

    async def get_by_run_and_key(
        self, run_id: str, key: str, version: int = 1
    ) -> QualityMetric | None:
        """Get quality metric by run ID and key"""
        result = await self.session.execute(
            select(QualityMetric).where(
                and_(
                    QualityMetric.run_id == run_id,
                    QualityMetric.key == key,
                    QualityMetric.version == version,
                )
            )
        )
        return result.scalar_one_or_none()

    async def list_by_run(self, run_id: str) -> list[QualityMetric]:
        """List quality metrics by run"""
        result = await self.session.execute(
            select(QualityMetric)
            .where(QualityMetric.run_id == run_id)
            .order_by(QualityMetric.computed_at.desc())
        )
        return list(result.scalars().all())

    async def list_by_batch(
        self,
        batch_id: str,
        key: str | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> tuple[list[QualityMetric], str | None]:
        """List quality metrics by batch with optional key filter"""
        query = (
            select(QualityMetric)
            .join(Run, QualityMetric.run_id == Run.run_id)
            .where(Run.batch_id == batch_id)
        )

        if key:
            query = query.where(QualityMetric.key == key)

        query = query.order_by(QualityMetric.computed_at.desc())

        if cursor:
            cursor_qm = await self.session.execute(
                select(QualityMetric).where(QualityMetric.qm_id == cursor)
            )
            cursor_obj = cursor_qm.scalar_one_or_none()
            if cursor_obj:
                query = query.where(QualityMetric.computed_at < cursor_obj.computed_at)

        query = query.limit(limit + 1)
        result = await self.session.execute(query)
        metrics = list(result.scalars().all())

        next_cursor = None
        if len(metrics) > limit:
            metrics = metrics[:limit]
            next_cursor = metrics[-1].qm_id

        return metrics, next_cursor

    async def delete_by_run(self, run_id: str) -> int:
        """Delete all quality metrics for a run"""
        result = await self.session.execute(
            delete(QualityMetric).where(QualityMetric.run_id == run_id)
        )
        rowcount = getattr(result, "rowcount", 0)
        return int(rowcount or 0)

    async def list_by_aggregation(
        self,
        aggregation_id: str,
        key: str | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> tuple[list[QualityMetric], str | None]:
        """List quality metrics for runs in an aggregation"""
        query = (
            select(QualityMetric)
            .join(
                AggregationMember,
                QualityMetric.run_id == AggregationMember.run_id,
            )
            .where(AggregationMember.aggregation_id == aggregation_id)
        )

        if key:
            query = query.where(QualityMetric.key == key)

        query = query.order_by(QualityMetric.computed_at.desc())

        if cursor:
            cursor_qm = await self.session.execute(
                select(QualityMetric).where(QualityMetric.qm_id == cursor)
            )
            cursor_obj = cursor_qm.scalar_one_or_none()
            if cursor_obj:
                query = query.where(QualityMetric.computed_at < cursor_obj.computed_at)

        query = query.limit(limit + 1)
        result = await self.session.execute(query)
        metrics = list(result.scalars().all())

        next_cursor = None
        if len(metrics) > limit:
            metrics = metrics[:limit]
            next_cursor = metrics[-1].qm_id

        return metrics, next_cursor



"""Pagination schemas"""

from typing import TypeVar

from pydantic import BaseModel, Field

T = TypeVar("T")


class PaginationParams(BaseModel):
    """Pagination parameters"""

    limit: int = Field(default=20, ge=1, le=100)
    cursor: str | None = None


class CursorPaginatedResponse[T](BaseModel):
    """Cursor-based paginated response"""

    items: list[T]
    next_cursor: str | None = None
    has_more: bool = False



"""Object Storage base interface"""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any


@dataclass
class ObjectRef:
    """Reference to an object in storage"""

    key: str
    bucket: str
    provider: str
    size: int | None = None
    content_type: str | None = None
    metadata: dict[str, str] | None = None


class ObjectStore(ABC):
    """Abstract base class for object storage implementations"""

    @abstractmethod
    async def put_bytes(
        self,
        key: str,
        data: bytes,
        content_type: str | None = None,
        metadata: dict[str, str] | None = None,
    ) -> ObjectRef:
        """Store bytes at the given key"""
        ...

    @abstractmethod
    async def put_json(
        self,
        key: str,
        data: Any,
        metadata: dict[str, str] | None = None,
    ) -> ObjectRef:
        """Store JSON data at the given key"""
        ...

    @abstractmethod
    async def get_bytes(self, key: str) -> bytes:
        """Retrieve bytes from the given key"""
        ...

    @abstractmethod
    async def get_json(self, key: str) -> Any:
        """Retrieve JSON data from the given key"""
        ...

    @abstractmethod
    async def exists(self, key: str) -> bool:
        """Check if an object exists at the given key"""
        ...

    @abstractmethod
    async def list_keys(self, prefix: str) -> list[ObjectRef]:
        """List objects with the given prefix"""
        ...

    @abstractmethod
    async def delete(self, key: str) -> bool:
        """Delete the object at the given key"""
        ...

    @abstractmethod
    async def presign_get(self, key: str, expires_sec: int = 3600) -> str:
        """Generate a presigned URL for GET operations"""
        ...

    @abstractmethod
    async def presign_put(
        self, key: str, content_type: str | None = None, expires_sec: int = 3600
    ) -> str:
        """Generate a presigned URL for PUT operations"""
        ...



"""Organization service."""

from __future__ import annotations

from axion_lab.schemas import CursorPaginatedResponse, OrgCreate, OrgResponse
from axion_lab.services._http import HttpTransport


class OrgService:
    """CRUD operations for organizations."""

    def __init__(self, http: HttpTransport) -> None:
        self._http = http

    def create(self, name: str) -> OrgResponse:
        """Create a new organization."""
        data = self._http.post("/orgs", json=OrgCreate(name=name).model_dump())
        return OrgResponse.model_validate(data)

    def list(
        self,
        *,
        limit: int = 20,
        cursor: str | None = None,
    ) -> CursorPaginatedResponse[OrgResponse]:
        """List organizations with cursor pagination."""
        params: dict[str, object] = {"limit": limit}
        if cursor:
            params["cursor"] = cursor
        data = self._http.get("/orgs", params=params)
        return CursorPaginatedResponse[OrgResponse].model_validate(data)

    def get(self, org_id: str) -> OrgResponse:
        """Get an organization by ID."""
        data = self._http.get(f"/orgs/{org_id}")
        return OrgResponse.model_validate(data)



"""Quality Metric service."""

from __future__ import annotations

from axion_lab.schemas import CursorPaginatedResponse, QualityMetricResponse
from axion_lab.services._http import HttpTransport


class QualityMetricService:
    """Read-only access to quality metrics (computed via DP)."""

    def __init__(self, http: HttpTransport) -> None:
        self._http = http

    def list_by_run(
        self,
        run_id: str,
        *,
        limit: int = 100,
        cursor: str | None = None,
    ) -> CursorPaginatedResponse[QualityMetricResponse]:
        """List quality metrics for a run."""
        params: dict[str, object] = {"limit": limit}
        if cursor:
            params["cursor"] = cursor
        data = self._http.get(f"/runs/{run_id}/quality-metrics", params=params)
        return CursorPaginatedResponse[QualityMetricResponse].model_validate(data)

    def list_by_batch(
        self,
        batch_id: str,
        *,
        key: str | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> CursorPaginatedResponse[QualityMetricResponse]:
        """List quality metrics for all runs in a batch."""
        params: dict[str, object] = {"limit": limit}
        if key:
            params["key"] = key
        if cursor:
            params["cursor"] = cursor
        data = self._http.get(f"/batches/{batch_id}/quality-metrics", params=params)
        return CursorPaginatedResponse[QualityMetricResponse].model_validate(data)



"""Kernel package - app startup and cross-cutting infrastructure"""

from axion_lab_server.shared.kernel.database import (
    close_db,
    create_engine,
    get_db,
    get_engine,
    get_session,
    get_session_maker,
)

__all__ = [
    "create_engine",
    "get_engine",
    "get_session_maker",
    "get_session",
    "get_db",
    "close_db",
]



import os
from logging.config import fileConfig

from alembic import context
from dotenv import load_dotenv
from sqlalchemy import create_engine, pool

# Load environment variables from .env file
load_dotenv()

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
if config.config_file_name is not None:
    fileConfig(config.config_file_name)

from axion_lab_server.repos.models import (  # noqa: F401 - ensure all models are imported
    entities,
)

# Import models for autogenerate support
from axion_lab_server.repos.models.base import Base

target_metadata = Base.metadata


def get_url() -> str:
    """Get database URL from environment or config."""
    url = os.getenv("DATABASE_URL")
    if url:
        # Convert async URL to sync URL for Alembic
        return url.replace("+asyncpg", "").replace("+aiosqlite", "")
    return config.get_main_option("sqlalchemy.url", "")


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = get_url()
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online() -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    url = get_url()
    connectable = create_engine(url, poolclass=pool.NullPool)

    with connectable.connect() as connection:
        context.configure(connection=connection, target_metadata=target_metadata)

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



"""Run Pin repository"""

from sqlalchemy import and_, delete, select
from sqlalchemy.ext.asyncio import AsyncSession

from axion_lab_server.repos.models.entities import RunPin
from axion_lab_server.shared.domain import PinType
from axion_lab_server.shared.libs import generate_id, utc_now


class RunPinRepository:
    """Repository for RunPin operations"""

    def __init__(self, session: AsyncSession):
        self.session = session

    async def create(
        self,
        run_id: str,
        batch_id: str,
        pin_type: PinType,
        pinned_by: str | None = None,
    ) -> RunPin:
        """Create a new pin (with upsert for champion)"""
        # If champion, remove existing champion first
        if pin_type == PinType.CHAMPION:
            await self.delete_champion(batch_id)

        pin = RunPin(
            pin_id=generate_id(),
            run_id=run_id,
            batch_id=batch_id,
            pin_type=pin_type.value,
            pinned_by=pinned_by,
            pinned_at=utc_now(),
        )
        self.session.add(pin)
        await self.session.flush()
        return pin

    async def get_by_run_and_type(
        self, run_id: str, pin_type: PinType
    ) -> RunPin | None:
        """Get pin by run ID and type"""
        result = await self.session.execute(
            select(RunPin).where(
                and_(
                    RunPin.run_id == run_id,
                    RunPin.pin_type == pin_type.value,
                )
            )
        )
        return result.scalar_one_or_none()

    async def delete_by_run_and_type(self, run_id: str, pin_type: PinType) -> bool:
        """Delete pin by run ID and type"""
        result = await self.session.execute(
            delete(RunPin).where(
                and_(
                    RunPin.run_id == run_id,
                    RunPin.pin_type == pin_type.value,
                )
            )
        )
        rowcount = getattr(result, "rowcount", 0)
        return int(rowcount or 0) > 0

    async def delete_champion(self, batch_id: str) -> bool:
        """Delete champion pin for a batch"""
        result = await self.session.execute(
            delete(RunPin).where(
                and_(
                    RunPin.batch_id == batch_id,
                    RunPin.pin_type == PinType.CHAMPION.value,
                )
            )
        )
        rowcount = getattr(result, "rowcount", 0)
        return int(rowcount or 0) > 0

    async def get_champion_run_id(self, batch_id: str) -> str | None:
        """Get champion run ID for a batch"""
        result = await self.session.execute(
            select(RunPin.run_id).where(
                and_(
                    RunPin.batch_id == batch_id,
                    RunPin.pin_type == PinType.CHAMPION.value,
                )
            )
        )
        return result.scalar_one_or_none()

    async def list_by_batch(self, batch_id: str) -> list[RunPin]:
        """List all pins for a batch"""
        result = await self.session.execute(
            select(RunPin)
            .where(RunPin.batch_id == batch_id)
            .order_by(RunPin.pinned_at.desc())
        )
        return list(result.scalars().all())



"""axion-lab server - API and database components

New module structure:
- shared/kernel: Database engine, session management
- shared/libs: Configuration and utilities
- shared/domain: Pydantic schemas (request/response models)
- repos: SQLAlchemy repositories and models
- gateways: External service integrations (object storage)
- ops: Background operations (DP runner)
- apps: FastAPI application and API routers
"""

__version__ = "0.1.0"



"""Local filesystem object storage implementation"""

import json
from pathlib import Path
from typing import Any
from urllib.parse import quote

from axion_lab_server.gateways.storage.base import ObjectRef, ObjectStore


class LocalObjectStore(ObjectStore):
    """Local filesystem implementation of ObjectStore"""

    def __init__(self, base_path: str, bucket: str = "local"):
        self.base_path = Path(base_path)
        self.bucket = bucket
        self.base_path.mkdir(parents=True, exist_ok=True)

    def _get_full_path(self, key: str) -> Path:
        """Get full filesystem path for a key"""
        return self.base_path / key

    async def put_bytes(
        self,
        key: str,
        data: bytes,
        content_type: str | None = None,
        metadata: dict[str, str] | None = None,
    ) -> ObjectRef:
        """Store bytes at the given key"""
        path = self._get_full_path(key)
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_bytes(data)

        # Store metadata in a sidecar file
        if metadata or content_type:
            meta = metadata or {}
            if content_type:
                meta["_content_type"] = content_type
            meta_path = Path(str(path) + ".meta")
            meta_path.write_text(json.dumps(meta))

        return ObjectRef(
            key=key,
            bucket=self.bucket,
            provider="local",
            size=len(data),
            content_type=content_type,
            metadata=metadata,
        )

    async def put_json(
        self,
        key: str,
        data: Any,
        metadata: dict[str, str] | None = None,
    ) -> ObjectRef:
        """Store JSON data at the given key"""
        json_bytes = json.dumps(data, ensure_ascii=False, indent=2).encode("utf-8")
        return await self.put_bytes(
            key, json_bytes, content_type="application/json", metadata=metadata
        )

    async def get_bytes(self, key: str) -> bytes:
        """Retrieve bytes from the given key"""
        path = self._get_full_path(key)
        if not path.exists():
            raise FileNotFoundError(f"Object not found: {key}")
        return path.read_bytes()

    async def get_json(self, key: str) -> Any:
        """Retrieve JSON data from the given key"""
        data = await self.get_bytes(key)
        return json.loads(data.decode("utf-8"))

    async def exists(self, key: str) -> bool:
        """Check if an object exists at the given key"""
        return self._get_full_path(key).exists()

    async def list_keys(self, prefix: str) -> list[ObjectRef]:
        """List objects with the given prefix"""
        prefix_path = self._get_full_path(prefix)
        results: list[ObjectRef] = []

        if prefix_path.is_dir():
            for path in prefix_path.rglob("*"):
                if path.is_file() and not path.name.endswith(".meta"):
                    rel_key = str(path.relative_to(self.base_path)).replace("\\", "/")
                    results.append(
                        ObjectRef(
                            key=rel_key,
                            bucket=self.bucket,
                            provider="local",
                            size=path.stat().st_size,
                        )
                    )
        elif prefix_path.exists():
            results.append(
                ObjectRef(
                    key=prefix,
                    bucket=self.bucket,
                    provider="local",
                    size=prefix_path.stat().st_size,
                )
            )

        return results

    async def delete(self, key: str) -> bool:
        """Delete the object at the given key"""
        path = self._get_full_path(key)
        if path.exists():
            path.unlink()
            # Also delete metadata file if exists
            meta_path = Path(str(path) + ".meta")
            if meta_path.exists():
                meta_path.unlink()
            return True
        return False

    async def presign_get(self, key: str, expires_sec: int = 3600) -> str:
        """Generate a file:// URL for local files"""
        path = self._get_full_path(key)
        return f"file://{quote(str(path.absolute()))}"

    async def presign_put(
        self, key: str, content_type: str | None = None, expires_sec: int = 3600
    ) -> str:
        """Generate a file:// URL for local files"""
        path = self._get_full_path(key)
        return f"file://{quote(str(path.absolute()))}"



from datetime import UTC, datetime
from types import SimpleNamespace
from unittest.mock import AsyncMock, Mock

import pytest
from fastapi import HTTPException

from axion_lab_server.apps.api.routers.artifacts import (
    _build_artifact_response,
    create_artifact,
    delete_artifact,
    list_artifacts,
)
from axion_lab_server.shared.domain import ArtifactCreate, ArtifactKind


def _artifact(*, artifact_id: str = "a1", run_id: str = "r1", meta_json: str = ""):
    return SimpleNamespace(
        artifact_id=artifact_id,
        run_id=run_id,
        kind=ArtifactKind.INLINE_JSON,
        type="evaluation",
        label="summary",
        meta_json=meta_json,
        created_at=datetime(2026, 2, 12, tzinfo=UTC),
    )


def test_build_artifact_response_parses_payload_and_meta() -> None:
    artifact = _artifact(meta_json='{"score": 0.91}')
    repo = Mock()
    repo.get_payload.return_value = {"result": "ok"}

    response = _build_artifact_response(artifact, repo)

    assert response.artifact_id == "a1"
    assert response.run_id == "r1"
    assert response.payload == {"result": "ok"}
    assert response.meta == {"score": 0.91}


@pytest.mark.asyncio
async def test_create_artifact_returns_built_response() -> None:
    run = SimpleNamespace(run_id="run-123")
    artifact = _artifact(artifact_id="art-1", run_id="run-123")
    repo = Mock()
    repo.create = AsyncMock(return_value=artifact)
    repo.get_payload.return_value = {"message": "created"}

    data = ArtifactCreate(
        kind=ArtifactKind.INLINE_JSON,
        type="evaluation",
        label="result",
        payload={"input": 1},
        meta={"source": "test"},
    )

    response = await create_artifact(data=data, run=run, repo=repo)

    repo.create.assert_awaited_once_with("run-123", data)
    assert response.artifact_id == "art-1"
    assert response.run_id == "run-123"
    assert response.payload == {"message": "created"}


@pytest.mark.asyncio
async def test_list_artifacts_returns_cursor_page() -> None:
    run = SimpleNamespace(run_id="run-321")
    artifacts = [
        _artifact(artifact_id="a1", run_id="run-321"),
        _artifact(artifact_id="a2", run_id="run-321"),
    ]

    repo = Mock()
    repo.list_by_run = AsyncMock(return_value=(artifacts, "next-cursor"))
    repo.get_payload.side_effect = [{"idx": 1}, {"idx": 2}]

    page = await list_artifacts(run=run, repo=repo, limit=2, cursor="cur-1")

    repo.list_by_run.assert_awaited_once_with("run-321", limit=2, cursor="cur-1")
    assert page.has_more is True
    assert page.next_cursor == "next-cursor"
    assert [item.artifact_id for item in page.items] == ["a1", "a2"]


@pytest.mark.asyncio
async def test_delete_artifact_returns_none_on_success() -> None:
    repo = Mock()
    repo.delete = AsyncMock(return_value=True)

    result = await delete_artifact(artifact_id="a1", repo=repo)

    repo.delete.assert_awaited_once_with("a1")
    assert result is None


@pytest.mark.asyncio
async def test_delete_artifact_raises_404_when_missing() -> None:
    repo = Mock()
    repo.delete = AsyncMock(return_value=False)

    with pytest.raises(HTTPException) as exc_info:
        await delete_artifact(artifact_id="missing", repo=repo)

    assert exc_info.value.status_code == 404
    assert exc_info.value.detail == "Artifact not found: missing"



"""Organization API router"""

from fastapi import APIRouter, status

from axion_lab_server.apps.api.deps import OrgPath, OrgRepo
from axion_lab_server.shared.domain import CursorPaginatedResponse, OrgCreate, OrgResponse

router = APIRouter(prefix="/orgs", tags=["Organizations"])


@router.post("", response_model=OrgResponse, status_code=status.HTTP_201_CREATED)
async def create_org(data: OrgCreate, repo: OrgRepo) -> OrgResponse:
    """Create a new organization"""
    org = await repo.create(data)
    return OrgResponse.model_validate(org)


@router.get("", response_model=CursorPaginatedResponse[OrgResponse])
async def list_orgs(
    repo: OrgRepo,
    limit: int = 20,
    cursor: str | None = None,
) -> CursorPaginatedResponse[OrgResponse]:
    """List all organizations with cursor pagination"""
    orgs, next_cursor = await repo.list_all(limit=limit, cursor=cursor)
    return CursorPaginatedResponse(
        items=[OrgResponse.model_validate(o) for o in orgs],
        next_cursor=next_cursor,
        has_more=next_cursor is not None,
    )


@router.get("/{org_id}", response_model=OrgResponse)
async def get_org(org: OrgPath) -> OrgResponse:
    """Get organization by ID"""
    return OrgResponse.model_validate(org)



"""Artifact API router"""

import json
from typing import Any

from fastapi import APIRouter, HTTPException, status

from axion_lab_server.apps.api.deps import ArtifactRepo, RunPath
from axion_lab_server.shared.domain import (
    ArtifactCreate,
    ArtifactResponse,
    CursorPaginatedResponse,
)

router = APIRouter(tags=["Artifacts"])


def _build_artifact_response(artifact: Any, repo: ArtifactRepo) -> ArtifactResponse:
    """Build artifact response with proper payload"""
    payload = repo.get_payload(artifact)
    meta = json.loads(artifact.meta_json) if artifact.meta_json else {}

    return ArtifactResponse(
        artifactId=artifact.artifact_id,
        runId=artifact.run_id,
        kind=artifact.kind,
        type=artifact.type,
        label=artifact.label,
        payload=payload,
        meta=meta,
        createdAt=artifact.created_at,
    )


@router.post(
    "/runs/{run_id}/artifacts",
    response_model=ArtifactResponse,
    status_code=status.HTTP_201_CREATED,
)
async def create_artifact(
    data: ArtifactCreate,
    run: RunPath,
    repo: ArtifactRepo,
) -> ArtifactResponse:
    """Create a new artifact for a run"""
    artifact = await repo.create(run.run_id, data)
    return _build_artifact_response(artifact, repo)


@router.get(
    "/runs/{run_id}/artifacts",
    response_model=CursorPaginatedResponse[ArtifactResponse],
)
async def list_artifacts(
    run: RunPath,
    repo: ArtifactRepo,
    limit: int = 100,
    cursor: str | None = None,
) -> CursorPaginatedResponse[ArtifactResponse]:
    """List artifacts for a run"""
    artifacts, next_cursor = await repo.list_by_run(
        run.run_id, limit=limit, cursor=cursor
    )
    return CursorPaginatedResponse(
        items=[_build_artifact_response(a, repo) for a in artifacts],
        next_cursor=next_cursor,
        has_more=next_cursor is not None,
    )


@router.delete(
    "/artifacts/{artifact_id}",
    status_code=status.HTTP_204_NO_CONTENT,
)
async def delete_artifact(
    artifact_id: str,
    repo: ArtifactRepo,
) -> None:
    """Delete an artifact"""
    deleted = await repo.delete(artifact_id)
    if not deleted:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Artifact not found: {artifact_id}",
        )



"""Axion Lab - Artifact-first Experiment Evaluation System"""

__version__ = "0.1.0"

from axion_lab.client import AxionLabClient
from axion_lab.high_level import (
    create_artifact,
    create_run,
    get_active_run_id,
    set_active_run,
)
from axion_lab.services._http import AxionLabHTTPError

__all__ = [
    "AxionLabClient",
    "AxionLabHTTPError",
    "create_run",
    "create_artifact",
    "get_active_run_id",
    "set_active_run",
]
